
/*
   lvm.c : Logical Volume Manager for Linux
              Copyright (C) 1998 Ingo Molnar

   LVM management functions.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
   
   You should have received a copy of the GNU General Public License
   (for example /usr/src/linux/COPYING); if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  
*/

#include <linux/module.h>
#include <linux/raid/md.h>
#include <linux/raid/lvm.h>

#define MAJOR_NR LVM_MAJOR
#define LVM_DRIVER

#include <linux/blk.h>

lvdev_t lv_devs[MAX_LVM_DEV];

static LVM_LIST_HEAD(lvm_managed_devices);

static lvm_rdisk_t * find_rdisk_all (kdev_t dev)
{
	struct lvm_list_head *tmp;
	lvm_rdisk_t *rdisk;

	tmp = lvm_managed_devices.next;
	while (tmp != &lvm_managed_devices) {
		rdisk = lvm_list_entry(tmp, lvm_rdisk_t, lvm_managed);
		if (rdisk->dev == dev)
			return rdisk;
		tmp = tmp->next;
	}
	return NULL;
}

/*
 * Prevent the device from being mounted or otherwise reused
 * in the LVM (or any other kernel subsystem), by adding a
 * dummy inode.
 */
static void lock_rdisk (lvm_rdisk_t *rdisk)
{
	/*
	 * Insert dummy inode to 'lock' the device
	 */
	rdisk->inode = get_empty_inode();
 	/*
	 * we dont care about other fields
	 */
	rdisk->inode->i_dev = rdisk->dev;
	insert_inode_hash(rdisk->inode);
}

static void unlock_rdisk (lvm_rdisk_t *rdisk)
{
	clear_inode(rdisk->inode);
}

#define NO_SB KERN_ERR \
"lvm: skipping device %s, could not read superblock.\n"

#define BAD_MAGIC KERN_ERR \
"lvm: invalid LVM superblock magic on %s\n"

#define OUT_OF_MEM KERN_ALERT \
"lvm: out of memory.\n"

#define BAD_CSUM KERN_WARNING \
"lvm: invalid superblock checksum on %s\n"

static int alloc_rdisk_sb (lvm_rdisk_t * rdisk)
{
	if (rdisk->sb)
		LVM_BUG();

	rdisk->sb = (lvm_super_t *) __get_free_page(GFP_KERNEL);
	if (!rdisk->sb) {
		printk (OUT_OF_MEM);
		return -EINVAL;
	}
	lvm_clear_page((unsigned long)rdisk->sb);

	return 0;
}

static void free_rdisk_sb (lvm_rdisk_t * rdisk)
{
	if (rdisk->sb) {
		free_page((unsigned long) rdisk->sb);
		rdisk->sb = NULL;
		rdisk->size = 0;
	} else {
		if (!rdisk->faulty)
			LVM_BUG();
	}
}

static unsigned int calc_sb_csum (lvm_super_t * sb)
{
	unsigned int disk_csum, csum;

	disk_csum = sb->sb_csum;
	sb->sb_csum = 0;
	csum = csum_partial((void *)sb, LVM_SB_BYTES, 0);
	sb->sb_csum = disk_csum;
	return csum;
}

/*
 * Check one LVM superblock for generic plausibility
 */

static int check_rdisk_sb (lvm_rdisk_t * rdisk)
{
	lvm_super_t *sb;
	int ret = -EINVAL;

	sb = rdisk->sb;
	if (!sb) {
		LVM_BUG();
		goto abort;
	}

	if (sb->lvm_magic != LVM_SB_MAGIC) {
		printk (BAD_MAGIC, partition_name(rdisk->dev));
		goto abort;
	}

	if (calc_sb_csum(sb) != sb->sb_csum)
		printk(BAD_CSUM, partition_name(rdisk->dev));
	ret = 0;
abort:
	return ret;
}

static int read_rdisk_sb (lvm_rdisk_t * rdisk)
{
	int ret = -EINVAL;
	struct buffer_head *bh = NULL;
	kdev_t dev = rdisk->dev;
	lvm_super_t *sb;

	if (!rdisk->sb) {
		LVM_BUG();
		goto abort;
	}	
	
	/*
	 * Read the superblock, it's at the beginning of the disk
	 */
	set_blocksize (dev, LVM_SB_BYTES);
	bh = bread (dev, 0, LVM_SB_BYTES);

	if (bh) {
		sb = (lvm_super_t *) bh->b_data;
		memcpy (rdisk->sb, sb, LVM_SB_BYTES);
	} else {
		printk (NO_SB,partition_name(rdisk->dev));
		goto abort;
	}
	ret = 0;
abort:
	if (bh)
		brelse (bh);
	return ret;
}

#undef BAD_CSUM
#undef BAD_MAGIC
#undef OUT_OF_MEM
#undef NO_SB


#define GETBLK_FAILED KERN_ERR \
"lvm: getblk failed for device %s\n"

static int write_rdisk_sb(lvm_rdisk_t * rdisk)
{
	struct buffer_head *bh;
	kdev_t dev;
	lvm_super_t *sb;

	if (!rdisk->sb) {
		LVM_BUG();
		return -1;
	}
	if (rdisk->sb->lvm_magic != LVM_SB_MAGIC) {
		LVM_BUG();
		return -1;
	}

	dev = rdisk->dev;

	printk("(write) %s\n", partition_name(dev));
	set_blocksize(dev, LVM_SB_BYTES);
	bh = getblk(dev, 0, LVM_SB_BYTES);
	if (!bh) {
		printk(GETBLK_FAILED, partition_name(dev));
		return 1;
	}
	memset(bh->b_data,0,bh->b_size);
	sb = (lvm_super_t *) bh->b_data;
	memcpy(sb, rdisk->sb, LVM_SB_BYTES);

	mark_buffer_uptodate(bh, 1);
	mark_buffer_dirty(bh, 1);
	ll_rw_block(WRITE, 1, &bh);
	wait_on_buffer(bh);
	brelse(bh);
	fsync_dev(dev);
	return 0;
}
#undef GETBLK_FAILED KERN_ERR

/*
 * Import a device, sanity check the superblock
 */
static int lvm_import_device (kdev_t newdev)
{
	int err = 0;
	lvm_rdisk_t *rdisk;
	unsigned int size;

	if (find_rdisk_all(newdev))
		return -EEXIST;

	rdisk = (lvm_rdisk_t *) kmalloc(sizeof(*rdisk), GFP_KERNEL);
	if (!rdisk) {
		printk("could not alloc mem for %s!\n", partition_name(newdev));
		return -ENOMEM;
	}
	memset(rdisk, 0, sizeof(*rdisk));

	if (!fs_may_mount(newdev)) {
		printk("lvm: can not import %s, has active inodes!\n",
			partition_name(newdev));
		err = -EBUSY;
		goto abort_free;
	}

	if ((err = alloc_rdisk_sb(rdisk)))
		goto abort_free;

	rdisk->dev = newdev;
	lock_rdisk(rdisk);

	size = blk_size[MAJOR(newdev)][MINOR(newdev)];
	if (!size) {
		printk("lvm: %s has zero size, marking faulty!\n",
				partition_name(newdev));
		rdisk->faulty = 1;
		goto out_faulty;
	}

	if ((err = read_rdisk_sb(rdisk))) {
		printk("lvm: %s has zero size,marking faulty!\n",
				partition_name(newdev));
		rdisk->faulty = 1;
		goto out_faulty;
	}
	if ((err = check_rdisk_sb(rdisk))) {
		printk("lvm: %s has invalid sb, skipping!\n",
				partition_name(newdev));
		rdisk->faulty = 1;
		goto out_faulty;
	}

	rdisk->old_dev = MKDEV(rdisk->sb->this_major,
					rdisk->sb->this_minor);

out_faulty:
	lvm_list_add(&rdisk->lvm_managed, &lvm_managed_devices);
	if (rdisk->faulty && rdisk->sb)
		free_rdisk_sb(rdisk);
	return 0;

abort_free:
	kfree(rdisk);
	return err;
}

/*
 * Searches all registered partitions for LVM devices
 * at boot time.
 */
lvm__initfunc(void autodetect_lvm(void))
{
	struct gendisk *disk;
	int i;

	printk(KERN_INFO "autodetecting LVM devices\n");

	for (disk = gendisk_head ; disk ; disk = disk->next)
		for (i = 0; i < disk->max_p*disk->max_nr; i++) {
			kdev_t dev = MKDEV(disk->major,i);

			if (disk->part[i].type == LINUX_LVM_PARTITION) {
				printk(KERN_INFO "found raw LVM partition %s\n",
						partition_name(dev));
				lvm_import_device(dev);
			}
		}
}

static void do_lvm_request (void)
{
	printk(KERN_ALERT "Got LVM request, not good...");
	return;
}

int lvm_size[MAX_LVM_DEV] = {0, };
static struct hd_struct lvm_hd_struct[MAX_LVM_DEV];
static int lvm_blocksizes[MAX_LVM_DEV];
static int lvm_maxreadahead[MAX_LVM_DEV];

static void lvm_geninit (struct gendisk *);

static struct gendisk lvm_gendisk=
{
	LVM_MAJOR,
	"lvm",
	0,
	1,
	MAX_LVM_DEV,
	lvm_geninit,
	lvm_hd_struct,
	lvm_size,
	MAX_LVM_DEV,
	NULL,
	NULL
};

static void lvm_geninit (struct gendisk *gdisk)
{
	int i;

	for(i = 0; i < MAX_LVM_DEV; i++) {
		lvm_blocksizes[i] = 1024;
		lvm_maxreadahead[i] = 128;
		lvm_gendisk.part[i].start_sect = -1; /* avoid partition check */
		lvm_gendisk.part[i].nr_sects = 0;
	}

	blksize_size[LVM_MAJOR] = lvm_blocksizes;
	lvm_set_global_readahead(lvm_maxreadahead);
}

static int get_version (void * arg)
{
	lvm_version_t ver;

	ver.major = LVM_MAJOR_VERSION;
	ver.minor = LVM_MINOR_VERSION;
	ver.patchlevel = LVM_PATCHLEVEL_VERSION;

	if (md_copy_to_user(arg, &ver, sizeof(ver)))
		return -EFAULT;

	return 0;
}

static int lvm_ioctl (struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
	unsigned int minor;
	int err = 0;
	struct hd_geometry *loc = (struct hd_geometry *) arg;

	if (!lvm_capable_admin())
		return -EACCES;

	minor = MINOR(inode->i_rdev);
	if (minor >= MAX_LVM_DEV)
		return -EINVAL;

	switch (cmd)
	{
		case LVM_VERSION:
			err = get_version((void *)arg);
			break;
      
		case BLKGETSIZE:   /* Return device size */
			if (!arg) {
				err = -EINVAL;
				break;
			}
			err = lvm_put_user(lvm_hd_struct[
				MINOR(inode->i_rdev)].nr_sects, (long *) arg);
			break;

		case BLKFLSBUF:
			fsync_dev (inode->i_rdev);
			invalidate_buffers (inode->i_rdev);
			break;

		case BLKRASET:
			if (arg > 0xff)
				err = -EINVAL;
			read_ahead[MAJOR(inode->i_rdev)] = arg;
			break;
    
		case BLKRAGET:
			if (!arg)
				err = -EINVAL;
			err = lvm_put_user (read_ahead[
				MAJOR(inode->i_rdev)], (long *) arg);
			break;
	/*
	 * We have a problem here : there is no easy way to give a CHS
	 * virtual geometry. We currently pretend that we have a 2 heads
	 * 4 sectors (with a BIG number of cylinders...). This drives
	 * dosfs just mad... ;-)
	 */
		case HDIO_GETGEO:
			if (!loc) {
				err = -EINVAL;
				break;
			}
			err = lvm_put_user (2, (char *) &loc->heads);
			if (err)
				break;
			err = lvm_put_user (4, (char *) &loc->sectors);
			if (err)
				break;
			err = lvm_put_user (lvm_hd_struct[minor].nr_sects/8,
						(short *) &loc->cylinders);
			if (err)
				break;
			err = lvm_put_user (lvm_hd_struct[MINOR(inode->i_rdev)
				].start_sect, (long *) &loc->start);
			break;
    
		default:
			printk(KERN_WARNING "%s(pid %d) used obsolete LVM ioctl, upgrade your software to use new ictls.\n", current->comm, current->pid);
			err = -EINVAL;
	}
	return err;
}

#if LINUX_VERSION_CODE < LinuxVersionCode(2,1,0)

static int lvm_open (struct inode *inode, struct file *file)
{
	/*
	 * Always succeed
	 */
	return (0);
}

static void lvm_release (struct inode *inode, struct file *file)
{
	sync_dev(inode->i_rdev);
}


static int lvm_read (struct inode *inode, struct file *file,
						char *buf, int count)
{
	if (!kdev_to_lvdev(LVM_FILE_TO_INODE(file)->i_rdev)->running)
		return -ENXIO;

	return block_read (inode, file, buf, count);
}

static int lvm_write (struct inode *inode, struct file *file,
						const char *buf, int count)
{
	if (!kdev_to_lvdev(LVM_FILE_TO_INODE(file)->i_rdev)->running)
		return -ENXIO;

	return block_write (inode, file, buf, count);
}

static struct file_operations lvm_fops=
{
	NULL,
	lvm_read,
	lvm_write,
	NULL,
	NULL,
	lvm_ioctl,
	NULL,
	lvm_open,
	lvm_release,
	block_fsync
};

#else

static int lvm_open (struct inode *inode, struct file *file)
{
	/*
	 * Always succeed
	 */
	return (0);
}

static int lvm_release (struct inode *inode, struct file *file)
{
	sync_dev(inode->i_rdev);
	return 0;
}

static ssize_t lvm_read (struct file *file, char *buf, size_t count,
			loff_t *ppos)
{
	if (!kdev_to_lvdev(LVM_FILE_TO_INODE(file)->i_rdev)->running)
		return -ENXIO;

	return block_read(file, buf, count, ppos);
}

static ssize_t lvm_write (struct file *file, const char *buf,
			 size_t count, loff_t *ppos)
{
	if (!kdev_to_lvdev(LVM_FILE_TO_INODE(file)->i_rdev)->running)
		return -ENXIO;

	return block_write(file, buf, count, ppos);
}

static struct file_operations lvm_fops=
{
	NULL,
	lvm_read,
	lvm_write,
	NULL,
	NULL,
	lvm_ioctl,
	NULL,
	lvm_open,
	NULL,
	lvm_release,
	block_fsync
};

#endif

static void extend_index_space (lvdev_t *lv)
{
#warning not yet
}

static inline unsigned int __get_free_block (pvdev_t *pv)
{
	if (pv->first_free_block == pv->max_blocks) {
		printk("lvm: pv %s full!\n", partition_name(pv->rdev->dev));
		return 0;
	}
	return pv->first_free_block++;
}

static unsigned int get_free_block (lvdev_t *lv)
{
	unsigned int idx;

	while (!lv->free_indices)
		extend_index_space(lv);

	idx = __get_free_block(&lv->pv);
	if (!idx)
		LVM_BUG();

	return idx;
}

static inline lvm_lptr_t * __get_free_index (lvdev_t *lv)
{
#warning not yet
	return 0;
}

static lvm_lptr_t * get_free_index (lvdev_t *lv)
{
	lvm_lptr_t *index;

	if (!lv->free_indices)
		extend_index_space(lv);

	index = __get_free_index(lv);
	if (!index)
		LVM_BUG();

	return index;
}


static unsigned int index_to_idx (lvm_lptr_t *index)
{
#warning not yet
	return 0;
}

static void bind_block_to_index (lvm_lptr_t *index,
					unsigned int block, lvdev_t *lv)
{
	index->data.phys_block = block;
}

unsigned int alloc_index (kdev_t dev)
{
	lvdev_t *lv = kdev_to_lvdev(dev);
	lvm_lptr_t *index;
	unsigned int block;

	index = get_free_index(lv);
	if (!index) {
		LVM_BUG();
		return 0;
	}
	block = get_free_block(lv);
	bind_block_to_index(index,block,lv);

	return index_to_idx(index);
}

#define INDEX_BITS 8
#define INDEX_DEPTH (32/INDEX_BITS)
#define INDEX_MASK ((1<<INDEX_BITS) - 1)

lvm_lptr_t * find_index (lvdev_t *lv, unsigned int lblock)
{
	lvm_lptr_t * index = lv->root_index;
	unsigned int idx, l;

	for (l = 0; l < INDEX_DEPTH; l++) {
		idx = lblock & INDEX_MASK;
		index += idx;
		index = (lvm_lptr_t *) (index->cpu_addr);
		if (!index) {
			return NULL;
		}
		lblock >>= INDEX_BITS;
	}
	return index;
}

static int lvm_map (kdev_t dev, kdev_t *rdev,
			unsigned long *rsector, int bsize_bits)
{
	lvdev_t *lv = kdev_to_lvdev(dev);
	lvm_lptr_t *index;
	unsigned int lblock = *rsector >> bsize_bits;
	int err = -EIO;

	index = find_index(lv, lblock);
	if (!index) {
		printk("no block %u yet ...\n", lblock);
		goto out;
	}

	err = 0;
out:
	return err;
}

lvm__initfunc(void lvm_init(void))
{
	printk (KERN_INFO "Logical Volume Manager for Linux %d.%d.%d\n",
		LVM_MAJOR_VERSION, LVM_MINOR_VERSION, LVM_PATCHLEVEL_VERSION);
	if (register_blkdev (LVM_MAJOR, "lvm", &lvm_fops))
	{
		printk (KERN_ALERT "Unable to get LVM major %d\n", LVM_MAJOR);
		return;
	}

	blk_dev[LVM_MAJOR].request_fn = DEVICE_REQUEST;
	blk_dev[LVM_MAJOR].current_request = NULL;
	read_ahead[LVM_MAJOR] = INT_MAX;
	lvm_gendisk.next = gendisk_head;

	gendisk_head = &lvm_gendisk;
}

void lvm_print_devices (void)
{
	struct lvm_list_head *tmp;
	lvm_rdisk_t *rdisk;

	printk("\n");
	printk("       **********************************\n");
	printk("       * <COMPLETE LVM STATE PRINTOUT> *\n");
	printk("       **********************************\n");

	tmp = lvm_managed_devices.next;
	while (tmp != &lvm_managed_devices) {
		rdisk = lvm_list_entry(tmp, lvm_rdisk_t, lvm_managed);
		printk("LVM device: <%s>\n", partition_name(rdisk->dev));
		tmp = tmp->next;
	}

	printk("       **********************************\n");
	printk("\n");
}

/*
 * This Linus-trick catches bugs via the linker.
 */

extern void __BUG__in__lvm_dot_c_1(void);
extern void __BUG__in__lvm_dot_c_2(void);
extern void __BUG__in__lvm_dot_c_3(void);
extern void __BUG__in__lvm_dot_c_4(void);
extern void __BUG__in__lvm_dot_c_5(void);
 
void bugcatcher (void)
{
        if (sizeof(lvm_block_group_t) != LVM_BLOCKSIZE)
                __BUG__in__lvm_dot_c_1();
        if (sizeof(lvm_index_block_t) != LVM_BLOCKSIZE)
                __BUG__in__lvm_dot_c_2();
        if (sizeof(lv_header_t) != LVM_BLOCKSIZE)
                __BUG__in__lvm_dot_c_3();
        if (sizeof(lvm_super_t) != LVM_BLOCKSIZE)
                __BUG__in__lvm_dot_c_4();
	if (sizeof(lvm_lptr_t) != 16)
                __BUG__in__lvm_dot_c_5();
}

