/*
 * xor.c : Multiple Devices driver for Linux
 *
 * Copyright (C) 1996, 1997, 1998, 1999 Ingo Molnar, Matti Aarnio, Jakub Jelinek
 *
 *
 * optimized RAID-5 checksumming functions.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * You should have received a copy of the GNU General Public License
 * (for example /usr/src/linux/COPYING); if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include <linux/module.h>
#include <linux/raid/md.h>
#ifdef __sparc_v9__
#include <asm/head.h>
#include <asm/asi.h>
#include <asm/visasm.h>
#endif

/*
 * we use the 'XOR function template' to register multiple xor
 * functions runtime. The kernel measures their speed upon bootup
 * and decides which one to use. (compile-time registration is
 * not enough as certain CPU features like MMX can only be detected
 * runtime)
 *
 * this architecture makes it pretty easy to add new routines
 * that are faster on certain CPUs, without killing other CPU's
 * 'native' routine. Although the current routines are belived
 * to be the physically fastest ones on all CPUs tested, but
 * feel free to prove me wrong and add yet another routine =B-)
 * --mingo
 */

#define SRCDEST (struct buffer_head *dest, struct buffer_head *source)

typedef void (*xor_block_t) SRCDEST;
xor_block_t xor_block = NULL;

#ifndef __sparc_v9__

struct xor_block_template;

struct xor_block_template {
	char * name;
	xor_block_t xor_block;
	int speed;
	struct xor_block_template * next;
};

struct xor_block_template * xor_functions = NULL;

#define XORBLOCK_TEMPLATE(x) \
static void xor_block_##x SRCDEST; \
static struct xor_block_template t_xor_block_##x = \
				 { #x, xor_block_##x, 0, NULL }; \
static void xor_block_##x SRCDEST

#ifdef __i386__
/*
 * high-speed RAID5 checksumming functions utilizing MMX instructions
 * Copyright (C) 1998 Ingo Molnar
 */

XORBLOCK_TEMPLATE(pII_mmx)
{
	int len  = dest->b_size;
	long *destp   = (long *) dest->b_data;
	long *sourcep = (long *) source->b_data;
	char fpu_save[108];
        int lines = (len>>7) - 1;

	if (!(current->flags & PF_USEDFPU))
		__asm__ __volatile__ ( " clts;\n");

	__asm__ __volatile__ ( " fsave %0; fwait\n"::"m"(fpu_save[0]) );

        __asm__ __volatile__ (

#define LD(x,y) \
        "       movq   8*("#x")(%%esi), %%mm"#y"   ;\n"

#define XO(x,y) \
        "       pxor   8*("#x")(%%edi), %%mm"#y"   ;\n"

#define ST(x,y) \
        "       movq %%mm"#y",   8*("#x")(%%esi)   ;\n"

#define BLOCK0(i) \
	ST(i,0)						\
					LD(i+4,4)	\
					XO(i+4,4)	\
		ST(i+1,1)				\
	LD(i+5,5)					\
	XO(i+5,5)					\
			ST(i+2,2)			\
		LD(i+6,6)				\
		XO(i+6,6)				\
				ST(i+3,3)		\
			LD(i+7,7)			\
			XO(i+7,7)

#define BLOCK1(i) \
	ST(i,4)						\
					LD(i+4,0)	\
					XO(i+4,0)	\
		ST(i+1,5)				\
	LD(i+5,1)					\
	XO(i+5,1)					\
			ST(i+2,6)			\
		LD(i+6,2)				\
		XO(i+6,2)				\
				ST(i+3,7)		\
			LD(i+7,3)			\
			XO(i+7,3)

	LD(0,0)
	XO(0,0)
		LD(1,1)
		XO(1,1)
			LD(2,2)
			XO(2,2)
				LD(3,3)
				XO(3,3)

	" .align 32,0x90		;\n"
        " 1:                            ;\n"

		BLOCK0(0)
		BLOCK1(4)
		BLOCK0(8)
		BLOCK1(12)

        "       addl $128, %%esi         ;\n"
        "       addl $128, %%edi         ;\n"
        "       decl %%ecx              ;\n"
        "       jnz 1b                  ;\n"

		BLOCK0(0)
		BLOCK1(4)
		BLOCK0(8)

	ST(12,4)
		ST(13,5)
			ST(14,6)
				ST(15,7)
        :
        : "D" (sourcep), "S" (destp), "c" (lines)
        : "memory" );

	__asm__ __volatile__ ( " frstor %0;\n"::"m"(fpu_save[0]) );

	if (!(current->flags & PF_USEDFPU))
		stts();
}

#undef LD
#undef XO
#undef ST

XORBLOCK_TEMPLATE(p5_mmx)
{
	int len  = dest->b_size;
	long *destp   = (long *) dest->b_data;
	long *sourcep = (long *) source->b_data;
	char fpu_save[108];
        int lines = len / 64;

	if (!(current->flags & PF_USEDFPU))
		__asm__ __volatile__ ( " clts;\n");

	__asm__ __volatile__ ( " fsave %0; fwait\n"::"m"(fpu_save[0]) );

        __asm__ __volatile__ (

        " .align 32,0x90                ;\n"
        " 1:                            ;\n"
        "       movq   (%%esi), %%mm0   ;\n"
        "       movq  8(%%esi), %%mm1   ;\n"
        "       pxor   (%%edi), %%mm0   ;\n"
        "       movq 16(%%esi), %%mm2   ;\n"
        "       movq %%mm0,   (%%esi)   ;\n"
        "       pxor  8(%%edi), %%mm1   ;\n"
        "       movq 24(%%esi), %%mm3   ;\n"
        "       movq %%mm1,  8(%%esi)   ;\n"
        "       pxor 16(%%edi), %%mm2   ;\n"
        "       movq 32(%%esi), %%mm4   ;\n"
        "       movq %%mm2, 16(%%esi)   ;\n"
        "       pxor 24(%%edi), %%mm3   ;\n"
        "       movq 40(%%esi), %%mm5   ;\n"
        "       movq %%mm3, 24(%%esi)   ;\n"
        "       pxor 32(%%edi), %%mm4   ;\n"
        "       movq 48(%%esi), %%mm6   ;\n"
        "       movq %%mm4, 32(%%esi)   ;\n"
        "       pxor 40(%%edi), %%mm5   ;\n"
        "       movq 56(%%esi), %%mm7   ;\n"
        "       movq %%mm5, 40(%%esi)   ;\n"
        "       pxor 48(%%edi), %%mm6   ;\n"
        "       pxor 56(%%edi), %%mm7   ;\n"
        "       movq %%mm6, 48(%%esi)   ;\n"
        "       movq %%mm7, 56(%%esi)   ;\n"
        
        "       addl $64, %%esi         ;\n"
        "       addl $64, %%edi         ;\n"
        "       decl %%ecx              ;\n"
        "       jnz 1b                  ;\n"

        : 
        : "D" (sourcep), "S" (destp), "c" (lines)
        : "memory" );

	__asm__ __volatile__ ( " frstor %0;\n"::"m"(fpu_save[0]) );

	if (!(current->flags & PF_USEDFPU))
		stts();
}
#endif /* __i386__ */
#endif /* !__sparc_v9__ */

#ifdef __sparc_v9__
/*
 * High speed xor_block operation for RAID4/5 utilizing the
 * UltraSparc Visual Instruction Set.
 *
 * Copyright (C) 1997, 1999 Jakub Jelinek (jj@ultra.linux.cz)
 *
 *	Requirements:
 *	!(((long)destp | (long)sourcep) & (64 - 1)) &&
 *	!(len & 127) && len >= 256
 *
 * It is done in pure assembly, as otherwise gcc makes it
 * a non-leaf function, which is not what we want.
 * Also, we don't measure the speeds as on other architectures,
 * as the measuring routine does not take into account cold caches
 * and the fact that xor_block_VIS bypasses the caches.
 * xor_block_32regs might be 5% faster if caches are hot and things
 * just right, but I think it is better not to pollute the caches.
 * Actually, if I'd just fight for speed for hot caches, I could
 * write a hybrid VIS/integer routine, which would do always two
 * 64B blocks in VIS and two in IEUs, but I really care more about
 * caches.
 */
extern void *VISenter(void);
extern void xor_block_VIS SRCDEST;

void __xor_block_VIS(void)
{
__asm__ ("
	.globl xor_block_VIS
xor_block_VIS:
	ldx	[%%o1 + %0], %%o1
	ldx	[%%o0 + %1], %%o2
	ldx	[%%o0 + %0], %%o0
	rd	%%fprs, %%o5
	andcc	%%o5, %2, %%g0
	be,pt	%%icc, 297f
	 sethi	%%hi(%5), %%g1
	jmpl	%%g1 + %%lo(%5), %%g7
	 add	%%g7, 8, %%g7
297:	wr	%%g0, %4, %%fprs
	wr	%%g0, %3, %%asi
	membar	#LoadStore|#StoreLoad|#StoreStore
	sub	%%o2, 128, %%o2
	ldda	[%%o0] %3, %%f0
	ldda	[%%o1] %3, %%f16
1:	ldda	[%%o0 + 64] %%asi, %%f32
	fxor	%%f0, %%f16, %%f16
	fxor	%%f2, %%f18, %%f18
	fxor	%%f4, %%f20, %%f20
	fxor	%%f6, %%f22, %%f22
	fxor	%%f8, %%f24, %%f24
	fxor	%%f10, %%f26, %%f26
	fxor	%%f12, %%f28, %%f28
	fxor	%%f14, %%f30, %%f30
	stda	%%f16, [%%o0] %3
	ldda	[%%o1 + 64] %%asi, %%f48
	ldda	[%%o0 + 128] %%asi, %%f0
	fxor	%%f32, %%f48, %%f48
	fxor	%%f34, %%f50, %%f50
	add	%%o0, 128, %%o0
	fxor	%%f36, %%f52, %%f52
	add	%%o1, 128, %%o1
	fxor	%%f38, %%f54, %%f54
	subcc	%%o2, 128, %%o2
	fxor	%%f40, %%f56, %%f56
	fxor	%%f42, %%f58, %%f58
	fxor	%%f44, %%f60, %%f60
	fxor	%%f46, %%f62, %%f62
	stda	%%f48, [%%o0 - 64] %%asi
	bne,pt	%%xcc, 1b
	 ldda	[%%o1] %3, %%f16
	ldda	[%%o0 + 64] %%asi, %%f32
	fxor	%%f0, %%f16, %%f16
	fxor	%%f2, %%f18, %%f18
	fxor	%%f4, %%f20, %%f20
	fxor	%%f6, %%f22, %%f22
	fxor	%%f8, %%f24, %%f24
	fxor	%%f10, %%f26, %%f26
	fxor	%%f12, %%f28, %%f28
	fxor	%%f14, %%f30, %%f30
	stda	%%f16, [%%o0] %3
	ldda	[%%o1 + 64] %%asi, %%f48
	membar	#Sync
	fxor	%%f32, %%f48, %%f48
	fxor	%%f34, %%f50, %%f50
	fxor	%%f36, %%f52, %%f52
	fxor	%%f38, %%f54, %%f54
	fxor	%%f40, %%f56, %%f56
	fxor	%%f42, %%f58, %%f58
	fxor	%%f44, %%f60, %%f60
	fxor	%%f46, %%f62, %%f62
	stda	%%f48, [%%o0 + 64] %%asi
	membar	#Sync|#StoreStore|#StoreLoad
	retl
	 wr	%%g0, 0, %%fprs
	" : :
	"i" (&((struct buffer_head *)0)->b_data),
	"i" (&((struct buffer_head *)0)->b_size),
	"i" (FPRS_FEF|FPRS_DU), "i" (ASI_BLK_P),
	"i" (FPRS_FEF), "i" (VISenter));
}
#endif /* __sparc_v9__ */

#if defined(__sparc__) && !defined(__sparc_v9__)
/*
 * High speed xor_block operation for RAID4/5 utilizing the
 * ldd/std SPARC instructions.
 *
 * Copyright (C) 1999 Jakub Jelinek (jj@ultra.linux.cz)
 *
 */

XORBLOCK_TEMPLATE(SPARC)
{
	int size  = dest->b_size;
	int lines = size / (sizeof (long)) / 8, i;
	long *destp   = (long *) dest->b_data;
	long *sourcep = (long *) source->b_data;

	for (i = lines; i > 0; i--) {
	  __asm__ __volatile__("
	  ldd [%0 + 0x00], %%g2
	  ldd [%0 + 0x08], %%g4
	  ldd [%0 + 0x10], %%o0
	  ldd [%0 + 0x18], %%o2
	  ldd [%1 + 0x00], %%o4
	  ldd [%1 + 0x08], %%l0
	  ldd [%1 + 0x10], %%l2
	  ldd [%1 + 0x18], %%l4
	  xor %%g2, %%o4, %%g2
	  xor %%g3, %%o5, %%g3
	  xor %%g4, %%l0, %%g4
	  xor %%g5, %%l1, %%g5
	  xor %%o0, %%l2, %%o0
	  xor %%o1, %%l3, %%o1
	  xor %%o2, %%l4, %%o2
	  xor %%o3, %%l5, %%o3
	  std %%g2, [%0 + 0x00]
	  std %%g4, [%0 + 0x08]
	  std %%o0, [%0 + 0x10]
	  std %%o2, [%0 + 0x18]
	  " : : "r" (destp), "r" (sourcep) : "g2", "g3", "g4", "g5", "o0", 
	  "o1", "o2", "o3", "o4", "o5", "l0", "l1", "l2", "l3", "l4", "l5");

	  destp += 8;
	  sourcep += 8;
	}
}
#endif /* __sparc_v[78]__ */

#ifndef __sparc_v9__

/*
 * this one works reasonably on any x86 CPU
 * (send me an assembly version for inclusion if you can make it faster)
 *
 * this one is just as fast as written in pure assembly on x86.
 * the reason for this separate version is that the
 * fast open-coded xor routine "32reg" produces suboptimal code
 * on x86, due to lack of registers.
 */
XORBLOCK_TEMPLATE(8regs)
{
	int len  = dest->b_size;
	long *destp   = (long *) dest->b_data;
	long *sourcep = (long *) source->b_data;
	long lines = len / (sizeof (long)) / 8, i;

	for (i = lines; i > 0; i--) {
		*(destp + 0) ^= *(sourcep + 0);
		*(destp + 1) ^= *(sourcep + 1);
		*(destp + 2) ^= *(sourcep + 2);
		*(destp + 3) ^= *(sourcep + 3);
		*(destp + 4) ^= *(sourcep + 4);
		*(destp + 5) ^= *(sourcep + 5);
		*(destp + 6) ^= *(sourcep + 6);
		*(destp + 7) ^= *(sourcep + 7);
		destp += 8;
		sourcep += 8;
	}
}

/*
 * platform independent RAID5 checksum calculation, this should
 * be very fast on any platform that has a decent amount of
 * registers. (32 or more)
 */
XORBLOCK_TEMPLATE(32regs)
{
	int size  = dest->b_size;
	int lines = size / (sizeof (long)) / 8, i;
	long *destp   = (long *) dest->b_data;
	long *sourcep = (long *) source->b_data;

	for (i = lines; i > 0; i--) {

	  /* LOTS of registers available...
	     We do explicite loop-unrolling here for code which
	     favours RISC machines.  In fact this is almoast direct
	     RISC assembly on Alpha and SPARC :-)  */

	  register long d0, d1, d2, d3, d4, d5, d6, d7;

	  d0 = destp[0];	/* Pull the stuff into registers	*/
	  d1 = destp[1];	/*  ... in bursts, if possible.		*/
	  d2 = destp[2];
	  d3 = destp[3];
	  d4 = destp[4];
	  d5 = destp[5];
	  d6 = destp[6];
	  d7 = destp[7];
	  d0 ^= sourcep[0];
	  d1 ^= sourcep[1];
	  d2 ^= sourcep[2];
	  d3 ^= sourcep[3];
	  d4 ^= sourcep[4];
	  d5 ^= sourcep[5];
	  d6 ^= sourcep[6];
	  d7 ^= sourcep[7];

	  destp[0] = d0;	/* Store the result (in burts)		*/
	  destp[1] = d1;
	  destp[2] = d2;
	  destp[3] = d3;
	  destp[4] = d4;	/* Store the result (in burts)		*/
	  destp[5] = d5;
	  destp[6] = d6;
	  destp[7] = d7;

	  destp += 8;
	  sourcep += 8;
	}
}

/*
 * (the -6*32 shift factor colors the cache)
 */
#define SIZE (PAGE_SIZE-6*32)

static void xor_speed ( struct xor_block_template * func, 
	struct buffer_head *b1, struct buffer_head *b2)
{
	int speed;
	unsigned long now;
	int i, count, max;

	func->next = xor_functions;
	xor_functions = func;

	/*
	 * count the number of XORs done during a whole jiffy.
	 * calculate the speed of checksumming from this.
	 * (we use a 2-page allocation to have guaranteed
	 * color L1-cache layout)
	 */
	max = 0;
	for (i = 0; i < 5; i++) {
		now = jiffies;
		count = 0;
		while (jiffies == now) {
			mb();
			func->xor_block(b1,b2);
			mb();
			count++;
			mb();
		}
		if (count > max)
			max = count;
	}

	speed = max * (HZ*SIZE/1024);
	func->speed = speed;

	printk( "   %-10s: %5d.%03d MB/sec\n", func->name,
		speed / 1000, speed % 1000);
}

static inline void pick_fastest_function(void)
{
	struct xor_block_template *f, *fastest;

	fastest = xor_functions;
	for (f = fastest; f; f = f->next) {
		if (f->speed > fastest->speed)
			fastest = f;
	}
	xor_block = fastest->xor_block;

	printk( "using fastest function: %s (%d.%03d MB/sec)\n", fastest->name,
		fastest->speed / 1000, fastest->speed % 1000);
}
 

void calibrate_xor_block(void)
{
	struct buffer_head b1, b2;

	memset(&b1,0,sizeof(b1));
	b2 = b1;

	b1.b_data = (char *) md__get_free_pages(GFP_KERNEL,2);
	if (!b1.b_data)
		goto out;
	b2.b_data = b1.b_data + 2*PAGE_SIZE + SIZE;

	b1.b_size = SIZE;

	printk(KERN_INFO "raid5: measuring checksumming speed\n");

	sti(); /* should be safe */

#if defined(__sparc__) && !defined(__sparc_v9__)
	printk(KERN_INFO "raid5: using high-speed SPARC checksum routine\n");
	xor_speed(&t_xor_block_SPARC,&b1,&b2);
#endif

#if 1
#ifdef __i386__
	if (md_cpu_has_mmx()) {
		printk(KERN_INFO
			"raid5: using high-speed MMX checksum routine\n");
		xor_speed(&t_xor_block_pII_mmx,&b1,&b2);
		xor_speed(&t_xor_block_p5_mmx,&b1,&b2);
	}
#endif
#endif
	xor_speed(&t_xor_block_8regs,&b1,&b2);
	xor_speed(&t_xor_block_32regs,&b1,&b2);

	free_pages((unsigned long)b1.b_data,2);

out:
	pick_fastest_function();
}

#else /* __sparc_v9__ */

void calibrate_xor_block(void)
{
	printk(KERN_INFO "raid5: using high-speed VIS checksum routine\n");
	xor_block = xor_block_VIS;
}

#endif /* __sparc_v9__ */

MD_EXPORT_SYMBOL(xor_block);

